import asyncio
import os
import time

import dask
from dask.distributed import Client
from distributed.utils_test import inc
import pytest
import skein

from dask_yarn import YarnCluster
from dask_yarn.core import (
    _make_specification,
    _make_submit_specification,
    _make_scheduler_kwargs,
)
from .conftest import check_is_shutdown

try:
    import bokeh  # noqa

    bokeh_installed = True
except Exception:
    bokeh_installed = False


@pytest.mark.parametrize("deploy_mode", ["remote", "local"])
def test_basic(deploy_mode, skein_client, conda_env):
    with YarnCluster(
        environment=conda_env,
        deploy_mode=deploy_mode,
        worker_memory="512 MiB",
        scheduler_memory="512 MiB",
        name="test-basic",
        skein_client=skein_client,
        dashboard_address=":8787",
        port=8786,
        worker_options={"resources": {"FOO": "BAZ"}},
        worker_class="dask.distributed.Nanny",
    ) as cluster:
        # Smoketest repr

        if bokeh_installed:
            assert cluster.dashboard_link is not None
            if deploy_mode == "local":
                assert ":8787" in cluster.dashboard_link

        if deploy_mode == "local":
            assert ":8786" in cluster.scheduler_address

        # Scale up
        cluster.scale(2)

        with Client(cluster) as client:
            future = client.submit(inc, 10)
            assert future.result() == 11
            client.get_versions(check=True)
            resource_tags = client.run(lambda dask_worker: dask_worker.total_resources)
            assert {"FOO": "BAZ"} in list(resource_tags.values())

        # Check that 2 workers exist
        start = time.time()
        while len(cluster.workers()) != 2:
            time.sleep(0.1)
            assert time.time() < start + 30, "timeout cluster.scale(2)"

        # Scale down
        cluster.scale(1)

        start = time.time()
        while len(cluster.workers()) != 1:
            time.sleep(0.1)
            assert time.time() < start + 30, "timeout cluster.scale(1)"

    check_is_shutdown(skein_client, cluster.app_id)


def test_adapt(skein_client, conda_env):
    with YarnCluster(
        environment=conda_env,
        deploy_mode="local",
        worker_memory="256 MiB",
        name="test-adapt",
        skein_client=skein_client,
    ) as cluster:

        cluster.adapt()

        with Client(cluster) as client:
            future = client.submit(lambda x: x + 1, 10)
            result = future.result()
            assert result == 11

        start = time.time()
        while cluster.workers():
            time.sleep(0.1)
            assert time.time() < start + 30, "auto-scaledown timeout"

    check_is_shutdown(skein_client, cluster.app_id)


def test_from_specification(skein_client, conda_env, tmpdir):
    spec = _make_specification(
        environment=conda_env,
        worker_memory="512 MiB",
        scheduler_memory="512 MiB",
        name="dask-yarn-test-from-specification",
    )
    fn = os.path.join(str(tmpdir), "spec.yaml")
    with open(fn, "w") as f:
        f.write(spec.to_yaml())

    with YarnCluster.from_specification(fn, skein_client=skein_client) as cluster:
        with Client(cluster):
            pass

    check_is_shutdown(skein_client, cluster.app_id)


def test_from_specification_errors():
    bad_spec = skein.ApplicationSpec.from_yaml(
        """
        name: bad_spec
        services:
          bad:
            resources:
              memory: 1 GiB
              vcores: 1
            script: exit 1
        """
    )
    with pytest.raises(ValueError):
        YarnCluster.from_specification(bad_spec)

    with pytest.raises(TypeError):
        YarnCluster.from_specification(object())


def test_from_application_id(skein_client, conda_env):
    with YarnCluster(
        environment=conda_env,
        worker_memory="512 MiB",
        scheduler_memory="512 MiB",
        name="test-from-application-id",
        skein_client=skein_client,
    ) as cluster:

        # Connect to the application with the application id
        cluster2 = YarnCluster.from_application_id(cluster.app_id, skein_client)

        cluster2.scale(1)

        start = time.time()
        while len(cluster2.workers()) != 1:
            time.sleep(0.1)
            assert time.time() < start + 30, "timeout cluster.scale(1)"

        del cluster2

        # Cluster is still running, finalizer not run in cluster2
        assert len(cluster.workers()) == 1

    check_is_shutdown(skein_client, cluster.app_id)


def test_from_current(skein_client, conda_env, monkeypatch, tmpdir):
    # Not running in a container
    with pytest.raises(ValueError) as exc:
        YarnCluster.from_current()
    assert str(exc.value) == "Not running inside a container"

    with YarnCluster(
        environment=conda_env,
        worker_memory="512 MiB",
        scheduler_memory="512 MiB",
        name="test-from-current",
        skein_client=skein_client,
    ) as cluster:

        # Patch environment so it looks like a container
        container_id = "container_1526134340424_0012_01_000005"
        cont_dir = tmpdir.mkdir(container_id)
        with open(str(cont_dir.join(".skein.crt")), "wb") as fil:
            fil.write(skein_client.security._get_bytes("cert"))
        with open(str(cont_dir.join(".skein.pem")), "wb") as fil:
            fil.write(skein_client.security._get_bytes("key"))

        for key, val in [
            ("SKEIN_APPLICATION_ID", cluster.app_id),
            ("CONTAINER_ID", container_id),
            ("SKEIN_APPMASTER_ADDRESS", cluster.application_client.address),
            ("LOCAL_DIRS", str(tmpdir)),
        ]:
            monkeypatch.setenv(key, val)

        import skein.core

        monkeypatch.setattr(skein.core, "properties", skein.core.Properties())

        cluster2 = YarnCluster.from_current()
        assert cluster2.app_id == cluster.app_id
        assert cluster2.scheduler_address == cluster.scheduler_address

        # Smoketest method
        cluster2.scale(1)

        start = time.time()
        while len(cluster2.workers()) != 1:
            time.sleep(0.1)
            assert time.time() < start + 30, "timeout cluster.scale(1)"

        del cluster2

        # Cluster is still running, finalizer not run in cluster2
        assert len(cluster.workers()) == 1

    check_is_shutdown(skein_client, cluster.app_id)


@pytest.mark.parametrize("deploy_mode", ["remote", "local"])
def test_configuration(deploy_mode):
    config = {
        "yarn": {
            "environment": "myenv.tar.gz",
            "queue": "myqueue",
            "name": "dask-yarn-tests",
            "user": "alice",
            "tags": ["a", "b", "c"],
            "specification": None,
            "deploy-mode": deploy_mode,
            "worker": {
                "memory": "1234 MiB",
                "count": 1,
                "vcores": 1,
                "restarts": -1,
                "gpus": 2,
                "env": {"foo": "bar"},
                "worker_class": "abc",
                "worker_options": {"FOO": "BAZ"},
            },
            "scheduler": {"memory": "1234 MiB", "vcores": 1, "gpus": 0},
            "host": "0.0.0.0",
            "port": 8786,
            "dashboard_address": ":8787",
        }
    }

    with dask.config.set(config):
        spec = _make_specification()
        assert spec.name == "dask-yarn-tests"
        assert spec.user == "alice"
        assert spec.queue == "myqueue"
        assert spec.tags == {"a", "b", "c"}
        assert spec.services["dask.worker"].resources.memory == 1234
        assert spec.services["dask.worker"].resources.gpus == 2
        assert spec.services["dask.worker"].env == {
            "foo": "bar",
            "worker_class": "abc",
            "worker_options": "eyJGT08iOiAiQkFaIn0=",
        }

        if deploy_mode == "remote":
            assert spec.services["dask.scheduler"].resources.memory == 1234
            assert spec.services["dask.scheduler"].resources.gpus == 0
        else:
            assert "dask.scheduler" not in spec.services

        kwargs = _make_scheduler_kwargs()
        assert kwargs["host"] == "0.0.0.0"
        assert kwargs["port"] == 8786
        assert kwargs["dashboard_address"] == ":8787"


def test_configuration_full_specification(conda_env, tmpdir):
    spec = _make_specification(
        environment=conda_env,
        worker_memory="512 MiB",
        scheduler_memory="512 MiB",
        name="test-configuration-full-specification",
    )
    fn = os.path.join(str(tmpdir), "spec.yaml")
    with open(fn, "w") as f:
        f.write(spec.to_yaml())

    # path to full specification
    with dask.config.set({"yarn": {"specification": fn}}):
        spec = _make_specification()
        assert spec == spec

    # full specification inlined
    with dask.config.set({"yarn": {"specification": spec.to_dict()}}):
        spec = _make_specification()
        assert spec == spec

    # when overrides specified, ignores full specification
    override = {"specification": spec.to_dict()}
    override.update(dask.config.get("yarn"))
    override["name"] = "config-name"
    with dask.config.set({"yarn": override}):
        spec = _make_specification(environment="foo", name="test-name", n_workers=4)
        assert spec.services["dask.worker"].instances == 4
        assert spec.name == "test-name"


def test_make_specification_errors():
    with dask.config.set({"yarn.environment": None}):
        with pytest.raises(ValueError) as info:
            _make_specification()
        assert "conda-pack" in str(info.value)

    with pytest.raises(ValueError) as info:
        _make_specification(environment="foo.tar.gz", deploy_mode="unknown")
    assert "deploy_mode" in str(info.value)

    with pytest.raises(ValueError) as info:
        _make_specification(environment="foo.tar.gz", worker_memory=1234)
    assert "1234 MiB" in str(info.value)

    with pytest.raises(ValueError) as info:
        _make_specification(environment="foo.tar.gz", scheduler_memory=1234)
    assert "1234 MiB" in str(info.value)


@pytest.mark.parametrize(
    "env,path", [("conda://env-name", "env-name"), ("conda:///env/path", "/env/path")]
)
def test_environment_conda(env, path):
    spec = _make_submit_specification(
        "script.py", args=("foo", "bar"), deploy_mode="remote", environment=env
    )
    scheduler = spec.services["dask.scheduler"]
    assert not scheduler.files
    assert (
        scheduler.script == ("conda activate %s\ndask-yarn services scheduler") % path
    )
    worker = spec.services["dask.worker"]
    assert not worker.files
    assert worker.script == ("conda activate %s\ndask-yarn services worker") % path
    client = spec.services["dask.client"]
    assert set(client.files) == {"script.py"}
    assert (
        client.script
        == ("conda activate %s\ndask-yarn services client script.py foo bar") % path
    )


def test_environment_venv():
    env = "venv:///path/to/environment"
    path = "/path/to/environment"
    spec = _make_submit_specification(
        "script.py", args=("foo", "bar"), deploy_mode="remote", environment=env
    )
    scheduler = spec.services["dask.scheduler"]
    assert not scheduler.files
    assert (
        scheduler.script
        == ("source %s/bin/activate\ndask-yarn services scheduler") % path
    )
    worker = spec.services["dask.worker"]
    assert not worker.files
    assert worker.script == ("source %s/bin/activate\ndask-yarn services worker") % path
    client = spec.services["dask.client"]
    assert set(client.files) == {"script.py"}
    assert (
        client.script
        == ("source %s/bin/activate\ndask-yarn services client script.py foo bar")
        % path
    )


def test_environment_python_path():
    env = "python:///path/to/python"
    cmd = "/path/to/python -m dask_yarn.cli"
    spec = _make_submit_specification(
        "script.py", args=("foo", "bar"), deploy_mode="remote", environment=env
    )
    scheduler = spec.services["dask.scheduler"]
    assert not scheduler.files
    assert scheduler.script == "%s services scheduler" % cmd
    worker = spec.services["dask.worker"]
    assert not worker.files
    assert worker.script == "%s services worker" % cmd
    client = spec.services["dask.client"]
    assert set(client.files) == {"script.py"}
    assert client.script == "%s services client script.py foo bar" % cmd


def test_environment_archive():
    env = "env.tar.gz"
    spec = _make_submit_specification(
        "script.py", args=("foo", "bar"), deploy_mode="remote", environment=env
    )
    scheduler = spec.services["dask.scheduler"]
    assert set(scheduler.files) == {"environment"}
    assert scheduler.script == (
        "source environment/bin/activate\n"
        "environment/bin/python -m dask_yarn.cli services scheduler"
    )
    worker = spec.services["dask.worker"]
    assert set(worker.files) == {"environment"}
    assert worker.script == (
        "source environment/bin/activate\n"
        "environment/bin/python -m dask_yarn.cli services worker"
    )
    client = spec.services["dask.client"]
    assert set(client.files) == {"environment", "script.py"}
    assert client.script == (
        "source environment/bin/activate\n"
        "environment/bin/python -m dask_yarn.cli services client script.py foo bar"
    )


@pytest.mark.parametrize("deploy_mode", ["local", "remote"])
def test_make_submit_specification(deploy_mode):
    spec = _make_submit_specification(
        "../script.py",
        deploy_mode=deploy_mode,
        environment="myenv.tar.gz",
        name="test-name",
        user="alice",
        client_vcores=2,
        client_memory="2 GiB",
    )

    assert spec.name == "test-name"
    assert spec.user == "alice"
    if deploy_mode == "local":
        assert set(spec.services) == {"dask.worker"}
        assert "environment" in spec.services["dask.worker"].files
    else:
        client = spec.services["dask.client"]
        scheduler = spec.services["dask.scheduler"]
        assert client.files["environment"] == scheduler.files["environment"]
        assert client.files["script.py"].source.startswith("file:///")
        assert client.resources.memory == 2048
        assert client.resources.vcores == 2

    config = {
        "yarn.name": "dask-yarn-tests",
        "yarn.queue": "myqueue",
        "yarn.environment": "myenv.tar.gz",
        "yarn.client.memory": "1234 MiB",
        "yarn.client.vcores": 2,
        "yarn.client.env": {"foo": "bar"},
    }

    with dask.config.set(config):
        spec = _make_submit_specification("script.py", deploy_mode=deploy_mode)
        assert spec.name == "dask-yarn-tests"
        assert spec.queue == "myqueue"
        if deploy_mode == "remote":
            assert spec.services["dask.client"].resources.memory == 1234
            assert spec.services["dask.client"].resources.vcores == 2
            assert spec.services["dask.client"].env == {"foo": "bar"}


def test_environment_relative_paths():
    relpath = "path/to/foo.tar.gz"
    abspath = os.path.abspath(relpath)
    a = _make_specification(environment=relpath).to_dict()
    b = _make_specification(environment=abspath).to_dict()
    assert a == b


@pytest.mark.asyncio
@pytest.mark.parametrize("deploy_mode", ["remote", "local"])
async def test_basic_async(deploy_mode, skein_client, conda_env):
    async with YarnCluster(
        environment=conda_env,
        deploy_mode=deploy_mode,
        worker_memory="512 MiB",
        scheduler_memory="512 MiB",
        name="test-basic-async",
        skein_client=skein_client,
        asynchronous=True,
    ) as cluster:
        # Smoketest repr
        repr(cluster)

        if bokeh_installed:
            assert cluster.dashboard_link is not None

        # Scale up
        await cluster.scale(2)

        async with Client(cluster, asynchronous=True) as client:
            result = await client.submit(inc, 10)
            assert result == 11
            await client.get_versions(check=True)

        # Check that 2 workers exist
        start = time.time()
        while len(await cluster.workers()) != 2:
            await asyncio.sleep(0.1)
            assert time.time() < start + 30, "timeout cluster.scale(2)"

        # Scale down
        await cluster.scale(1)

        start = time.time()
        while len(await cluster.workers()) != 1:
            await asyncio.sleep(0.1)
            assert time.time() < start + 30, "timeout cluster.scale(1)"

    check_is_shutdown(skein_client, cluster.app_id)


def test_widget_and_html_reprs(skein_client, conda_env):
    pytest.importorskip("ipywidgets")

    with YarnCluster(
        environment=conda_env,
        deploy_mode="local",
        worker_memory="256 MiB",
        name="test-widget",
        skein_client=skein_client,
    ) as cluster:
        # Smoke test widget
        cluster._widget()

        # Test non-widget html repr
        assert cluster.app_id in cluster._repr_html_()

        assert "0" in cluster._widget_status()

        # Scale up and wait
        cluster.scale(1)
        start = time.time()
        while len(cluster._observed) != 1:
            time.sleep(0.1)
            assert time.time() < start + 30, "timeout cluster.scale(1)"

        assert "1" in cluster._widget_status()

        # Scale down
        cluster.scale(1)

    check_is_shutdown(skein_client, cluster.app_id)


def test_logs(conda_env, skein_client):
    with YarnCluster(
        environment=conda_env,
        deploy_mode="local",
        worker_memory="256 MiB",
        name="test-widget",
        skein_client=skein_client,
    ) as cluster:
        cluster.scale(2)
        start = time.time()
        while len(cluster._observed) != 2:
            time.sleep(0.1)
            assert time.time() < start + 30, "timeout cluster.scale(2)"

        logs = cluster.logs()
        assert len(logs) == 3

        logs = cluster.logs(scheduler=True, workers=False)
        assert len(logs) == 1

        logs = cluster.logs(scheduler=False, workers=False)
        assert len(logs) == 0

        logs = cluster.logs(scheduler=False, workers=True)
        assert len(logs) == 2

    check_is_shutdown(skein_client, cluster.app_id)



import pytest

import skein
import dask_yarn
from dask_yarn.cli import main
from .conftest import ensure_shutdown, wait_for_completion, get_logs


def run_command(command, error=False):
    with pytest.raises(SystemExit) as exc:
        main([arg for arg in command.split(" ") if arg])
    if error:
        assert exc.value.code != 0
    else:
        assert exc.value.code == 0


@pytest.mark.parametrize(
    "command",
    [
        "",
        "submit",
        "status",
        "kill",
        "services",
        "services scheduler",
        "services worker",
        "services client",
    ],
)
def test_cli_help(command, capfd):
    run_command(command + " -h")

    out, err = capfd.readouterr()
    assert not err
    assert "usage: dask-yarn" in out


@pytest.mark.parametrize("group", ["", "services"])
def test_cli_call_command_group(group, capfd):
    run_command(group, error=True)

    out, err = capfd.readouterr()
    assert not out
    assert "usage: dask-yarn" in err


def test_cli_version(capfd):
    run_command("--version")

    out, err = capfd.readouterr()
    assert not err
    assert dask_yarn.__version__ in out


GOOD_TEST_SCRIPT = """
import dask_yarn
from dask.distributed import Client

print("Connecting to cluster")
cluster = dask_yarn.YarnCluster.from_current()

print("Creating Dask Client")
client = Client(cluster)

print("Doing some computation")
assert client.submit(lambda x: x + 1, 1).result() == 2

print("Done!")
"""


BAD_TEST_SCRIPT = """
import dask_yarn
from dask.distributed import Client

print("Connecting to cluster")
cluster = dask_yarn.YarnCluster.from_current()

print("Creating Dask Client")
client = Client(cluster)

print("Intentionally Erroring")
assert False, "Failed!"
"""


scripts = {"good": GOOD_TEST_SCRIPT, "bad": BAD_TEST_SCRIPT}


@pytest.mark.parametrize(
    "script_kind, final_status, searchtxt",
    [("good", "SUCCEEDED", "Done!"), ("bad", "FAILED", "Failed!")],
)
def test_cli_submit_and_status(
    script_kind, final_status, searchtxt, tmpdir, conda_env, skein_client, capfd
):
    script_path = str(tmpdir.join("script.py"))
    with open(script_path, "w") as fil:
        fil.write(scripts[script_kind])

    run_command(
        "submit "
        "--name test-cli-submit-and-status "
        "--environment %s "
        "--worker-count 1 "
        "--worker-memory 256MiB "
        "--worker-vcores 1 "
        "--scheduler-memory 256MiB "
        "--scheduler-vcores 1 "
        "--client-memory 128MiB "
        "--client-vcores 1 "
        "%s" % (conda_env, script_path)
    )
    out, err = capfd.readouterr()
    # Logs go to err
    assert "INFO" in err
    app_id = out.strip()
    assert "\n" not in app_id

    with ensure_shutdown(skein_client, app_id, status=final_status):
        # Wait for app to start
        skein_client.connect(app_id)

        # `dask-yarn status`
        run_command("status %s" % app_id)
        out, err = capfd.readouterr()
        assert "INFO" in err
        assert len(out.splitlines()) == 2
        assert "RUNNING" in out

        wait_for_completion(skein_client, app_id, timeout=60)

    logs = get_logs(app_id)
    assert searchtxt in logs


def test_cli_submit_with_args(tmpdir, conda_env, skein_client, capfd):
    script = (
        "import sys\n"
        "args = sys.argv[1:]\n"
        'assert args == ["a", "b", "c"]\n'
        'print("Done!")'
    )

    script_path = str(tmpdir.join("script.py"))
    with open(script_path, "w") as fil:
        fil.write(script)

    run_command(
        "submit "
        "--name test-cli-submit-with-args "
        "--environment %s "
        "--worker-count 0 "
        "--scheduler-memory 256MiB "
        "--scheduler-vcores 1 "
        "--client-memory 128MiB "
        "--client-vcores 1 "
        "%s a b c" % (conda_env, script_path)
    )
    out, err = capfd.readouterr()
    # Logs go to err
    assert "INFO" in err
    app_id = out.strip()
    assert "\n" not in app_id

    with ensure_shutdown(skein_client, app_id, status="SUCCEEDED"):
        # Wait for app to start
        skein_client.connect(app_id)
        wait_for_completion(skein_client, app_id, timeout=60)

    logs = get_logs(app_id)
    assert "Done!" in logs


@pytest.mark.parametrize(
    "script_kind, final_status, searchtxt",
    [("good", "SUCCEEDED", "Done!"), ("bad", "FAILED", "Failed!")],
)
def test_cli_submit_local(
    script_kind, final_status, searchtxt, tmpdir, conda_env, skein_client, capfd
):
    if script_kind == "good":
        error = False
        script = GOOD_TEST_SCRIPT
    else:
        error = True
        script = BAD_TEST_SCRIPT

    script_path = str(tmpdir.join("script.py"))
    with open(script_path, "w") as fil:
        fil.write(script)

    run_command(
        "submit "
        "--name test-cli-submit-local "
        "--environment %s "
        "--deploy-mode local "
        "--worker-count 1 "
        "--worker-memory 256MiB "
        "--worker-vcores 1 "
        "--scheduler-memory 256MiB "
        "--scheduler-vcores 1 "
        "--client-memory 128MiB "
        "--client-vcores 1 "
        "%s" % (conda_env, script_path),
        error=error,
    )
    out, err = capfd.readouterr()
    # Logs go to err
    assert "INFO" in err
    assert searchtxt in out or searchtxt in err


@pytest.mark.parametrize(
    "deploy_mode, script_kind",
    [("remote", "good"), ("local", "good"), ("local", "bad")],
)
def test_cli_submit_temp_credentials(
    deploy_mode, script_kind, tmpdir, conda_env, skein_client, capfd
):
    if script_kind == "good":
        error = False
        final_status = "SUCCEEDED"
        searchtxt = "Done!"
        script = GOOD_TEST_SCRIPT
    else:
        error = True
        final_status = "FAILED"
        searchtxt = "Failed!"
        script = BAD_TEST_SCRIPT

    script_path = str(tmpdir.join("script.py"))
    with open(script_path, "w") as fil:
        fil.write(script)

    run_command(
        "submit "
        "--name test-cli-submit-temp-credentials "
        "--deploy-mode %s "
        "--environment %s "
        "--worker-count 1 "
        "--worker-memory 256MiB "
        "--worker-vcores 1 "
        "--scheduler-memory 256MiB "
        "--scheduler-vcores 1 "
        "--client-memory 128MiB "
        "--client-vcores 1 "
        "--temporary-security-credentials "
        "%s" % (deploy_mode, conda_env, script_path),
        error=error,
    )

    out, err = capfd.readouterr()
    # Logs go to err
    assert "INFO" in err

    if deploy_mode == "remote":
        app_id = out.strip()
        assert "\n" not in app_id

        with ensure_shutdown(skein_client, app_id, status=final_status):
            app = skein_client.connect(app_id)
            with pytest.raises(skein.ConnectionError):
                # Can't communicate due to mismatched security credentials
                app.get_specification()
            wait_for_completion(skein_client, app_id, timeout=60)

        logs = get_logs(app_id)
        assert searchtxt in logs
    else:
        assert searchtxt in out or searchtxt in err


def test_cli_kill(tmpdir, conda_env, skein_client, capfd):
    script_path = str(tmpdir.join("script.py"))
    with open(script_path, "w") as fil:
        fil.write(GOOD_TEST_SCRIPT)

    run_command(
        "submit "
        "--name test-cli-kill "
        "--environment %s "
        "--worker-count 1 "
        "--worker-memory 256MiB "
        "--worker-vcores 1 "
        "--scheduler-memory 256MiB "
        "--scheduler-vcores 1 "
        "--client-memory 128MiB "
        "--client-vcores 1 "
        "%s" % (conda_env, script_path)
    )
    out, err = capfd.readouterr()
    # Logs go to err
    assert "INFO" in err
    app_id = out.strip()
    assert "\n" not in app_id

    with ensure_shutdown(skein_client, app_id, status="KILLED"):
        # Wait for app to start
        skein_client.connect(app_id)

        # `dask-yarn kill`
        run_command("kill %s" % app_id)
        out, err = capfd.readouterr()
        assert "INFO" in err
        assert not out

        wait_for_completion(skein_client, app_id, timeout=60)



import argparse
import os
import base64
import shutil
import subprocess
import json
import sys
import tempfile
from contextlib import contextmanager
from urllib.parse import urlparse

import skein
from skein.utils import format_table, humanize_timedelta
from tornado.ioloop import IOLoop, TimeoutError
from distributed import Scheduler
from distributed.cli.utils import install_signal_handlers
from distributed.proctitle import (
    enable_proctitle_on_children,
    enable_proctitle_on_current,
)

from . import __version__
from .core import _make_submit_specification, YarnCluster, _get_skein_client


class _Formatter(argparse.HelpFormatter):
    """Format with a fixed argument width, due to bug in argparse measuring
    argument widths"""

    @property
    def _action_max_length(self):
        return 16

    @_action_max_length.setter
    def _action_max_length(self, value):
        pass

    def _format_args(self, action, default_metavar):
        """Format remainder arguments nicer"""
        get_metavar = self._metavar_formatter(action, default_metavar)
        if action.nargs == argparse.REMAINDER:
            return "[%s...]" % get_metavar(1)
        return super(_Formatter, self)._format_args(action, default_metavar)


class _VersionAction(argparse.Action):
    def __init__(
        self,
        option_strings,
        version=None,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
        help="Show version then exit",
    ):
        super(_VersionAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )
        self.version = version

    def __call__(self, parser, namespace, values, option_string=None):
        print(self.version % {"prog": parser.prog})
        sys.exit(0)


def fail(msg):
    print(msg, file=sys.stderr)
    sys.exit(1)


def add_help(parser):
    parser.add_argument(
        "--help", "-h", action="help", help="Show this help message then exit"
    )


def arg(*args, **kwargs):
    return (args, kwargs)


def subcommand(subparsers, name, help, *args):
    def _(func):
        parser = subparsers.add_parser(
            name,
            help=help,
            formatter_class=_Formatter,
            description=help,
            add_help=False,
        )
        parser.set_defaults(func=func)
        for arg in args:
            parser.add_argument(*arg[0], **arg[1])
        add_help(parser)
        func.parser = parser
        return func

    return _


def node(subs, name, help):
    @subcommand(subs, name, help)
    def f():
        fail(f.parser.format_usage())

    f.subs = f.parser.add_subparsers(metavar="command", dest="command")
    f.subs.required = True
    return f


entry = argparse.ArgumentParser(
    prog="dask-yarn",
    description="Deploy Dask on Apache YARN",
    formatter_class=_Formatter,
    add_help=False,
)
add_help(entry)
entry.add_argument(
    "--version",
    action=_VersionAction,
    version="%(prog)s " + __version__,
    help="Show version then exit",
)
entry.set_defaults(func=lambda: fail(entry.format_usage()))
entry_subs = entry.add_subparsers(metavar="command", dest="command")
entry_subs.required = True


def _parse_env(service, env):
    out = {}
    if env is None:
        return out
    for item in env:
        elements = item.split("=")
        if len(elements) != 2:
            raise ValueError("Invalid parameter to --%s-env: %r" % (service, env))
        key, val = elements
        out[key.strip()] = val.strip()
    return out


# Exposed for testing
def _parse_submit_kwargs(**kwargs):
    if kwargs.get("worker_env") is not None:
        kwargs["worker_env"] = _parse_env("worker", kwargs["worker_env"])
    if kwargs.get("client_env") is not None:
        kwargs["client_env"] = _parse_env("client", kwargs["client_env"])
    if kwargs.get("tags") is not None:
        kwargs["tags"] = set(map(str.strip, kwargs["tags"].split(",")))
    if kwargs.get("worker_count") is not None:
        kwargs["n_workers"] = kwargs.pop("worker_count")
    return kwargs


@subcommand(
    entry_subs,
    "submit",
    "Submit a Dask application to a YARN cluster",
    arg("script", help="Path to a python script to run on the client"),
    arg(
        "args",
        nargs=argparse.REMAINDER,
        help="Any additional arguments to forward to `script`",
    ),
    arg("--name", help="The application name"),
    arg("--queue", help="The queue to deploy to"),
    arg(
        "--user",
        help=(
            "The user to submit the application on behalf of. Default "
            "is the current user - submitting as a different user "
            "requires proxy-user permissions."
        ),
    ),
    arg(
        "--tags",
        help=(
            "A comma-separated list of strings to use as " "tags for this application."
        ),
    ),
    arg(
        "--environment",
        help=(
            "Path to the Python environment to use. See the docs "
            "for more information"
        ),
    ),
    arg(
        "--deploy-mode",
        help=(
            "Either 'remote' (default) or 'local'. If 'remote', the "
            "scheduler and client will be deployed in a YARN "
            "container. If 'local', they will be run locally."
        ),
    ),
    arg("--worker-count", type=int, help="The number of workers to initially start."),
    arg(
        "--worker-vcores",
        type=int,
        help="The number of virtual cores to allocate per worker.",
    ),
    arg(
        "--worker-memory",
        type=str,
        help=(
            "The amount of memory to allocate per worker. Accepts a "
            "unit suffix (e.g. '2 GiB' or '4096 MiB'). Will be "
            "rounded up to the nearest MiB."
        ),
    ),
    arg(
        "--worker-restarts",
        type=int,
        help=(
            "The maximum number of worker restarts to allow before "
            "failing the application. Default is unlimited."
        ),
    ),
    arg(
        "--worker-env",
        type=str,
        action="append",
        help=(
            "Environment variables to set on the workers. Pass a "
            "key-value pair like ``--worker-env key=val``. May "
            "be used more than once."
        ),
    ),
    arg(
        "--client-vcores",
        type=int,
        help="The number of virtual cores to allocate for the client.",
    ),
    arg(
        "--client-memory",
        type=str,
        help=(
            "The amount of memory to allocate for the client. "
            "Accepts a unit suffix (e.g. '2 GiB' or '4096 MiB'). "
            "Will be rounded up to the nearest MiB."
        ),
    ),
    arg(
        "--client-env",
        type=str,
        action="append",
        help=(
            "Environment variables to set on the client. Pass a "
            "key-value pair like ``--client-env key=val``. May "
            "be used more than once."
        ),
    ),
    arg(
        "--scheduler-vcores",
        type=int,
        help="The number of virtual cores to allocate for the scheduler.",
    ),
    arg(
        "--scheduler-memory",
        type=str,
        help=(
            "The amount of memory to allocate for the scheduler. "
            "Accepts a unit suffix (e.g. '2 GiB' or '4096 MiB'). "
            "Will be rounded up to the nearest MiB."
        ),
    ),
    arg(
        "--temporary-security-credentials",
        action="store_true",
        help=(
            "Instead of using a consistent set of TLS credentials "
            "for all clusters, create a fresh set just for this "
            "application."
        ),
    ),
)
def submit(script, args=None, temporary_security_credentials=False, **kwargs):
    kwargs = _parse_submit_kwargs(**kwargs)
    args = args or []
    spec = _make_submit_specification(script, args=args, **kwargs)

    if temporary_security_credentials:
        security = skein.Security.new_credentials()
    else:
        security = None

    skein_client = _get_skein_client(security=security)

    if "dask.scheduler" in spec.services:
        # deploy_mode == 'remote'
        app_id = skein_client.submit(spec)
        print(app_id)
    else:
        # deploy_mode == 'local'
        if not os.path.exists(script):
            raise ValueError("%r doesn't exist locally" % script)

        with maybe_tempdir(
            temporary_security_credentials
        ) as security_dir, YarnCluster.from_specification(
            spec, skein_client=skein_client
        ) as cluster:
            env = dict(os.environ)
            env.update(
                {
                    "DASK_APPLICATION_ID": cluster.app_id,
                    "DASK_APPMASTER_ADDRESS": cluster.application_client.address,
                }
            )
            if temporary_security_credentials:
                security.to_directory(security_dir)
                env["DASK_SECURITY_CREDENTIALS"] = security_dir

            retcode = subprocess.call([sys.executable, script] + args, env=env)

            if retcode == 0:
                cluster.shutdown("SUCCEEDED")
            else:
                cluster.shutdown(
                    "FAILED",
                    "Exception in submitted dask application, "
                    "see logs for more details",
                )
                sys.exit(retcode)


@contextmanager
def maybe_tempdir(create=False):
    """Contextmanager for consistent syntax for maybe creating a tempdir"""
    if create:
        try:
            path = tempfile.mkdtemp()
            yield path
        finally:
            shutil.rmtree(path)
    else:
        yield None


app_id = arg("app_id", help="The application id", metavar="APP_ID")


@subcommand(
    entry_subs, "status", "Check the status of a submitted Dask application", app_id
)
def status(app_id):
    report = _get_skein_client().application_report(app_id)
    header = [
        "application_id",
        "name",
        "state",
        "status",
        "containers",
        "vcores",
        "memory",
        "runtime",
    ]
    data = [
        (
            report.id,
            report.name,
            report.state,
            report.final_status,
            report.usage.num_used_containers,
            report.usage.used_resources.vcores,
            report.usage.used_resources.memory,
            humanize_timedelta(report.runtime),
        )
    ]
    print(format_table(header, data))


@subcommand(entry_subs, "kill", "Kill a Dask application", app_id)
def kill(app_id):
    _get_skein_client().kill_application(app_id)


services = node(entry_subs, "services", "Manage Dask services")


@subcommand(services.subs, "scheduler", "Start a Dask scheduler process")
def scheduler():  # pragma: nocover
    app_client = skein.ApplicationClient.from_current()

    enable_proctitle_on_current()
    enable_proctitle_on_children()

    if sys.platform.startswith("linux"):
        import resource  # module fails importing on Windows

        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        limit = max(soft, hard // 2)
        resource.setrlimit(resource.RLIMIT_NOFILE, (limit, hard))

    loop = IOLoop.current()
    scheduler = Scheduler(loop=loop, dashboard_address=("", 0))
    install_signal_handlers(loop)

    def post_addresses():
        # Set dask.dashboard before dask.scheduler since the YarnCluster object
        # waits on dask.scheduler only
        if "dashboard" in scheduler.services:
            bokeh_port = scheduler.services["dashboard"].port
            bokeh_host = urlparse(scheduler.address).hostname
            bokeh_address = "http://%s:%d" % (bokeh_host, bokeh_port)
            app_client.kv["dask.dashboard"] = bokeh_address.encode()
        app_client.kv["dask.scheduler"] = scheduler.address.encode()

    async def run():
        await scheduler
        await loop.run_in_executor(None, post_addresses)
        await scheduler.finished()

    try:
        loop.run_sync(run)
    except (KeyboardInterrupt, TimeoutError):
        pass
    finally:
        scheduler.stop()


@subcommand(
    services.subs,
    "worker",
    "Start a Dask worker process",
    arg(
        "--nthreads",
        type=int,
        help=("Number of threads. Defaults to number of vcores in " "container"),
    ),
    arg(
        "--memory_limit",
        type=str,
        help=(
            "Maximum memory available to the worker. This can be an "
            "integer (in bytes), a string (like '5 GiB' or '500 "
            "MiB'), or 0 (no memory management). Defaults to the "
            "container memory limit."
        ),
    ),
)
def worker(nthreads=None, memory_limit=None):  # pragma: nocover
    if memory_limit is None:
        memory_limit = int(skein.properties.container_resources.memory * 2 ** 20)
    if nthreads is None:
        nthreads = skein.properties.container_resources.vcores

    worker_options = dict(memory_limit=memory_limit, nthreads=nthreads)
    app_client = skein.ApplicationClient.from_current()

    scheduler = app_client.kv.wait("dask.scheduler").decode()
    worker_class = os.environ.get("worker_class", "dask.distributed.Nanny")
    options = os.environ.get("worker_options", str(base64.b64encode(b"{}")))

    # options come in as a string
    # encode string as bytes
    # decode base64
    # load with json
    options = base64.b64decode(options.encode())
    worker_options.update(json.loads(options))
    print(worker_options)
    command = [
        sys.executable,
        "-m",
        "distributed.cli.dask_spec",
        scheduler,
        "--spec",
        "%s"  # in yaml double single quotes escape the single quote
        % json.dumps(
            {
                "cls": worker_class,
                "opts": {"name": skein.properties.container_id, **worker_options},
            }
        ),
    ]
    print(command)
    subprocess.check_call(command)
    loop = IOLoop.current()

    install_signal_handlers(loop)


@subcommand(
    services.subs,
    "client",
    "Start a Dask client process",
    arg("script", help="Path to a Python script to run."),
    arg(
        "args",
        nargs=argparse.REMAINDER,
        help="Any additional arguments to forward to `script`",
    ),
)
def client(script, args=None):  # pragma: nocover
    app = skein.ApplicationClient.from_current()
    args = args or []

    if not os.path.exists(script):
        raise ValueError("%r doesn't exist" % script)

    retcode = subprocess.call([sys.executable, script] + args)

    if retcode == 0:
        app.shutdown("SUCCEEDED")
    else:
        print(
            "User submitted application %s failed with returncode "
            "%d, shutting down." % (script, retcode)
        )
        app.shutdown(
            "FAILED",
            "Exception in submitted dask application, " "see logs for more details",
        )


def main(args=None):
    kwargs = vars(entry.parse_args(args=args))
    kwargs.pop("command", None)  # Drop unnecessary `command` arg
    func = kwargs.pop("func")
    func(**kwargs)
    sys.exit(0)


if __name__ == "__main__":  # pragma: nocover
    main()



import asyncio
import math
import os
import base64
import sys
import json
import warnings
import weakref
from contextlib import contextmanager
from urllib.parse import urlparse

import dask
from distributed.core import rpc
from distributed.deploy.adaptive_core import AdaptiveCore
from distributed.scheduler import Scheduler
from distributed.utils import (
    format_bytes,
    log_errors,
    LoopRunner,
    format_dashboard_link,
    parse_timedelta,
    Log,
    Logs,
)

import skein
from skein.utils import cached_property


_memory_error = """Memory specification for the `{0}` take string parameters
with units like "4 GiB" or "2048 MiB"

You provided:       {1}
Perhaps you meant: "{1} MiB"
"""

_one_MiB = 2 ** 20


_widget_status_template = """
<div>
<style scoped>
    .dataframe tbody tr th:only-of-type {
        vertical-align: middle;
    }

    .dataframe tbody tr th {
        vertical-align: top;
    }

    .dataframe thead th {
        text-align: right;
    }
</style>
<table style="text-align: right;">
    <tr><th>Workers</th> <td>%d</td></tr>
    <tr><th>Cores</th> <td>%d</td></tr>
    <tr><th>Memory</th> <td>%s</td></tr>
</table>
</div>
"""


async def cancel_task(task):
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass


class DaskYarnError(skein.SkeinError):
    """An error involving the dask-yarn skein Application"""

    pass


@contextmanager
def submit_and_handle_failures(skein_client, spec):
    app_id = skein_client.submit(spec)
    try:
        try:
            yield skein_client.connect(app_id, security=spec.master.security)
        except BaseException:
            # Kill the application on any failures
            skein_client.kill_application(app_id)
            raise
    except (skein.ConnectionError, skein.ApplicationNotRunningError):
        # If the error was an application error, raise appropriately
        raise DaskYarnError(
            (
                "Failed to start dask-yarn {app_id}\n"
                "See the application logs for more information:\n\n"
                "$ yarn logs -applicationId {app_id}"
            ).format(app_id=app_id)
        )


def parse_memory(memory, field):
    if isinstance(memory, str):
        memory = dask.utils.parse_bytes(memory)
    elif memory < _one_MiB:
        raise ValueError(_memory_error.format(field, memory))
    return int(math.ceil(memory / _one_MiB))


def lookup(kwargs, a, b):
    return kwargs[a] if kwargs.get(a) is not None else dask.config.get(b)


def _get_skein_client(skein_client=None, security=None):
    if skein_client is None:
        # Silence warning about credentials not being written yet
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return skein.Client(security=security)
    return skein_client


def _files_and_build_script(environment):
    parsed = urlparse(environment)
    scheme = parsed.scheme

    if scheme in {"conda", "venv", "python"}:
        path = environment[len(scheme) + 3 :]
        files = {}
        if scheme == "conda":
            setup = "conda activate %s" % path
            cli = "dask-yarn"
        elif scheme == "venv":
            setup = "source %s/bin/activate" % path
            cli = "dask-yarn"
        else:
            setup = ""
            cli = "%s -m dask_yarn.cli" % path
    else:
        files = {"environment": environment}
        setup = "source environment/bin/activate"
        cli = "environment/bin/python -m dask_yarn.cli"

    def build_script(cmd):
        command = "%s %s" % (cli, cmd)
        return "\n".join([setup, command]) if setup else command

    return files, build_script


def _make_specification(**kwargs):
    """Create specification to run Dask Cluster

    This creates a ``skein.ApplicationSpec`` to run a dask cluster with the
    scheduler in a YARN container. See the docstring for ``YarnCluster`` for
    more details.
    """
    if all(v is None for v in kwargs.values()) and dask.config.get(
        "yarn.specification"
    ):
        # No overrides and full specification in configuration
        spec = dask.config.get("yarn.specification")
        if isinstance(spec, dict):
            return skein.ApplicationSpec.from_dict(spec)
        return skein.ApplicationSpec.from_file(spec)

    deploy_mode = lookup(kwargs, "deploy_mode", "yarn.deploy-mode")
    if deploy_mode not in {"remote", "local"}:
        raise ValueError(
            "`deploy_mode` must be one of {'remote', 'local'}, " "got %r" % deploy_mode
        )

    name = lookup(kwargs, "name", "yarn.name")
    queue = lookup(kwargs, "queue", "yarn.queue")
    tags = lookup(kwargs, "tags", "yarn.tags")
    user = lookup(kwargs, "user", "yarn.user")

    environment = lookup(kwargs, "environment", "yarn.environment")
    if environment is None:
        msg = (
            "You must provide a path to a Python environment for the workers.\n"
            "This may be one of the following:\n"
            "- A conda environment archived with conda-pack\n"
            "- A virtual environment archived with venv-pack\n"
            "- A path to a conda environment, specified as conda://...\n"
            "- A path to a virtual environment, specified as venv://...\n"
            "- A path to a python binary to use, specified as python://...\n"
            "\n"
            "See http://yarn.dask.org/environments.html for more information."
        )
        raise ValueError(msg)

    n_workers = lookup(kwargs, "n_workers", "yarn.worker.count")
    worker_restarts = lookup(kwargs, "worker_restarts", "yarn.worker.restarts")
    worker_env = lookup(kwargs, "worker_env", "yarn.worker.env")
    worker_class = lookup(kwargs, "worker_class", "yarn.worker.worker_class")
    worker_env["worker_class"] = worker_class

    worker_options = lookup(kwargs, "worker_options", "yarn.worker.worker_options")
    # json dump string
    # encode as base64
    # decode from bytes to string
    worker_options = base64.b64encode(json.dumps(worker_options).encode()).decode()
    worker_env["worker_options"] = worker_options
    worker_vcores = lookup(kwargs, "worker_vcores", "yarn.worker.vcores")
    worker_gpus = lookup(kwargs, "worker_gpus", "yarn.worker.gpus")
    worker_memory = parse_memory(
        lookup(kwargs, "worker_memory", "yarn.worker.memory"), "worker"
    )

    services = {}

    files, build_script = _files_and_build_script(environment)

    if deploy_mode == "remote":
        scheduler_vcores = lookup(kwargs, "scheduler_vcores", "yarn.scheduler.vcores")
        scheduler_gpus = lookup(kwargs, "scheduler_gpus", "yarn.scheduler.gpus")
        scheduler_memory = parse_memory(
            lookup(kwargs, "scheduler_memory", "yarn.scheduler.memory"), "scheduler"
        )

        services["dask.scheduler"] = skein.Service(
            instances=1,
            resources=skein.Resources(
                vcores=scheduler_vcores, memory=scheduler_memory, gpus=scheduler_gpus
            ),
            max_restarts=0,
            files=files,
            script=build_script("services scheduler"),
        )
        worker_depends = ["dask.scheduler"]
    else:
        worker_depends = None

    services["dask.worker"] = skein.Service(
        instances=n_workers,
        resources=skein.Resources(
            vcores=worker_vcores, memory=worker_memory, gpus=worker_gpus
        ),
        max_restarts=worker_restarts,
        depends=worker_depends,
        files=files,
        env=worker_env,
        script=build_script("services worker"),
    )

    spec = skein.ApplicationSpec(
        name=name, queue=queue, tags=tags, user=user, services=services
    )
    return spec


def _make_submit_specification(script, args=(), **kwargs):
    spec = _make_specification(**kwargs)

    environment = lookup(kwargs, "environment", "yarn.environment")
    files, build_script = _files_and_build_script(environment)

    if "dask.scheduler" in spec.services:
        # deploy_mode == 'remote'
        client_vcores = lookup(kwargs, "client_vcores", "yarn.client.vcores")
        client_memory = lookup(kwargs, "client_memory", "yarn.client.memory")
        client_gpus = lookup(kwargs, "client_gpus", "yarn.client.gpus")
        client_env = lookup(kwargs, "client_env", "yarn.client.env")
        client_memory = parse_memory(client_memory, "client")

        script_name = os.path.basename(script)
        files[script_name] = script

        spec.services["dask.client"] = skein.Service(
            instances=1,
            resources=skein.Resources(
                vcores=client_vcores, memory=client_memory, gpus=client_gpus
            ),
            max_restarts=0,
            depends=["dask.scheduler"],
            files=files,
            env=client_env,
            script=build_script(
                "services client %s %s" % (script_name, " ".join(args))
            ),
        )
    return spec


def _make_scheduler_kwargs(**kwargs):
    host = lookup(kwargs, "host", "yarn.host")
    port = lookup(kwargs, "port", "yarn.port")
    dashboard_address = lookup(kwargs, "dashboard_address", "yarn.dashboard-address")
    return {"host": host, "port": port, "dashboard_address": dashboard_address}


class YarnCluster(object):
    """Start a Dask cluster on YARN.

    You can define default values for this in Dask's ``yarn.yaml``
    configuration file. See http://docs.dask.org/en/latest/configuration.html
    for more information.

    Parameters
    ----------
    environment : str, optional
        The Python environment to use. Can be one of the following:

          - A path to an archived Python environment
          - A path to a conda environment, specified as `conda:///...`
          - A path to a virtual environment, specified as `venv:///...`
          - A path to a python executable, specifed as `python:///...`

        Note that if not an archive, the paths specified must be valid on all
        nodes in the cluster.
    n_workers : int, optional
        The number of workers to initially start.
    worker_vcores : int, optional
        The number of virtual cores to allocate per worker.
    worker_memory : str, optional
        The amount of memory to allocate per worker. Accepts a unit suffix
        (e.g. '2 GiB' or '4096 MiB'). Will be rounded up to the nearest MiB.
    worker_restarts : int, optional
        The maximum number of worker restarts to allow before failing the
        application. Default is unlimited.
    worker_env : dict, optional
        A mapping of environment variables to their values. These will be set
        in the worker containers before starting the dask workers.
    worker_gpus : int, options
        The number of gpus to allocate per worker
    scheduler_vcores : int, optional
        The number of virtual cores to allocate per scheduler.
    scheduler_gpus : int, options
        The number of gpus to allocate per scheduler
    scheduler_memory : str, optional
        The amount of memory to allocate to the scheduler. Accepts a unit
        suffix (e.g. '2 GiB' or '4096 MiB'). Will be rounded up to the nearest
        MiB.
    deploy_mode : {'remote', 'local'}, optional
        The deploy mode to use. If ``'remote'``, the scheduler will be deployed
        in a YARN container. If ``'local'``, the scheduler will run locally,
        which can be nice for debugging. Default is ``'remote'``.
    name : str, optional
        The application name.
    queue : str, optional
        The queue to deploy to.
    tags : sequence, optional
        A set of strings to use as tags for this application.
    user : str, optional
        The user to submit the application on behalf of. Default is the current
        user - submitting as a different user requires user permissions, see
        the YARN documentation for more information.
    host : str, optional
        Host address on which the scheduler will listen. Only used if
        ``deploy_mode='local'``. Defaults to ``'0.0.0.0'``.
    port : int, optional
        The port on which the scheduler will listen. Only used if
        ``deploy_mode='local'``. Defaults to ``0`` for a random port.
    dashboard_address : str
        Address on which to the dashboard server will listen. Only used if
        ``deploy_mode='local'``. Defaults to ':0' for a random port.
    skein_client : skein.Client, optional
        The ``skein.Client`` to use. If not provided, one will be started.
    asynchronous : bool, optional
        If true, starts the cluster in asynchronous mode, where it can be used
        in other async code.
    loop : IOLoop, optional
        The IOLoop instance to use. Defaults to the current loop in
        asynchronous mode, otherwise a background loop is started.

    Examples
    --------
    >>> cluster = YarnCluster(environment='my-env.tar.gz', ...)
    >>> cluster.scale(10)
    """

    def __init__(
        self,
        environment=None,
        n_workers=None,
        worker_vcores=None,
        worker_memory=None,
        worker_restarts=None,
        worker_env=None,
        worker_class=None,
        worker_options=None,
        worker_gpus=None,
        scheduler_vcores=None,
        scheduler_gpus=None,
        scheduler_memory=None,
        deploy_mode=None,
        name=None,
        queue=None,
        tags=None,
        user=None,
        host=None,
        port=None,
        dashboard_address=None,
        skein_client=None,
        asynchronous=False,
        loop=None,
    ):

        spec = _make_specification(
            environment=environment,
            n_workers=n_workers,
            worker_vcores=worker_vcores,
            worker_memory=worker_memory,
            worker_restarts=worker_restarts,
            worker_env=worker_env,
            worker_class=worker_class,
            worker_options=worker_options,
            worker_gpus=worker_gpus,
            scheduler_vcores=scheduler_vcores,
            scheduler_gpus=scheduler_gpus,
            scheduler_memory=scheduler_memory,
            deploy_mode=deploy_mode,
            name=name,
            queue=queue,
            tags=tags,
            user=user,
        )
        self._init_common(
            spec=spec,
            host=host,
            port=port,
            dashboard_address=dashboard_address,
            asynchronous=asynchronous,
            loop=loop,
            skein_client=skein_client,
        )

    @classmethod
    def from_specification(cls, spec, skein_client=None, asynchronous=False, loop=None):
        """Start a dask cluster from a skein specification.

        Parameters
        ----------
        spec : skein.ApplicationSpec, dict, or filename
            The application specification to use. Must define at least one
            service: ``'dask.worker'``. If no ``'dask.scheduler'`` service is
            defined, a scheduler will be started locally.
        skein_client : skein.Client, optional
            The ``skein.Client`` to use. If not provided, one will be started.
        asynchronous : bool, optional
            If true, starts the cluster in asynchronous mode, where it can be
            used in other async code.
        loop : IOLoop, optional
            The IOLoop instance to use. Defaults to the current loop in
            asynchronous mode, otherwise a background loop is started.
        """
        self = super(YarnCluster, cls).__new__(cls)
        if isinstance(spec, dict):
            spec = skein.ApplicationSpec.from_dict(spec)
        elif isinstance(spec, str):
            spec = skein.ApplicationSpec.from_file(spec)
        elif not isinstance(spec, skein.ApplicationSpec):
            raise TypeError(
                "spec must be an ApplicationSpec, dict, or path, "
                "got %r" % type(spec).__name__
            )
        if "dask.worker" not in spec.services:
            raise ValueError(
                "Provided Skein specification must include a 'dask.worker' service"
            )

        self._init_common(
            spec=spec, asynchronous=asynchronous, loop=loop, skein_client=skein_client
        )
        return self

    @classmethod
    def from_current(cls, asynchronous=False, loop=None):
        """Connect to an existing ``YarnCluster`` from inside the cluster.

        Parameters
        ----------
        asynchronous : bool, optional
            If true, starts the cluster in asynchronous mode, where it can be
            used in other async code.
        loop : IOLoop, optional
            The IOLoop instance to use. Defaults to the current loop in
            asynchronous mode, otherwise a background loop is started.

        Returns
        -------
        YarnCluster
        """
        self = super(YarnCluster, cls).__new__(cls)
        app_id = os.environ.get("DASK_APPLICATION_ID", None)
        app_address = os.environ.get("DASK_APPMASTER_ADDRESS", None)
        security_dir = os.environ.get("DASK_SECURITY_CREDENTIALS", None)
        if app_id is not None and app_address is not None:
            security = (
                None
                if security_dir is None
                else skein.Security.from_directory(security_dir)
            )
            app = skein.ApplicationClient(app_address, app_id, security=security)
        else:
            app = skein.ApplicationClient.from_current()

        self._init_common(application_client=app, asynchronous=asynchronous, loop=loop)
        return self

    @classmethod
    def from_application_id(
        cls, app_id, skein_client=None, asynchronous=False, loop=None
    ):
        """Connect to an existing ``YarnCluster`` with a given application id.

        Parameters
        ----------
        app_id : str
            The existing cluster's application id.
        skein_client : skein.Client
            The ``skein.Client`` to use. If not provided, one will be started.
        asynchronous : bool, optional
            If true, starts the cluster in asynchronous mode, where it can be
            used in other async code.
        loop : IOLoop, optional
            The IOLoop instance to use. Defaults to the current loop in
            asynchronous mode, otherwise a background loop is started.

        Returns
        -------
        YarnCluster
        """
        self = super(YarnCluster, cls).__new__(cls)
        skein_client = _get_skein_client(skein_client)
        app = skein_client.connect(app_id)

        self._init_common(
            application_client=app,
            asynchronous=asynchronous,
            loop=loop,
            skein_client=skein_client,
        )
        return self

    def _init_common(
        self,
        spec=None,
        application_client=None,
        host=None,
        port=None,
        dashboard_address=None,
        asynchronous=False,
        loop=None,
        skein_client=None,
    ):
        self.spec = spec
        self.application_client = application_client
        self._scheduler_kwargs = _make_scheduler_kwargs(
            host=host,
            port=port,
            dashboard_address=dashboard_address,
        )
        self._scheduler = None
        self.scheduler_info = {}
        self._requested = set()
        self.scheduler_comm = None
        self._watch_worker_status_task = None
        self._start_task = None
        self._stop_task = None
        self._finalizer = None
        self._adaptive = None
        self._adaptive_options = {}
        self._skein_client = skein_client
        self._asynchronous = asynchronous
        self._loop_runner = LoopRunner(loop=loop, asynchronous=asynchronous)
        self._loop_runner.start()

        if not self.asynchronous:
            self._sync(self._start_internal())

    def _start_cluster(self):
        """Start the cluster and initialize state"""

        skein_client = _get_skein_client(self._skein_client)

        if "dask.scheduler" not in self.spec.services:
            # deploy_mode == 'local'
            scheduler_address = self._scheduler.address
            for k in ["dashboard", "bokeh"]:
                if k in self._scheduler.services:
                    dashboard_port = self._scheduler.services[k].port
                    dashboard_host = urlparse(scheduler_address).hostname
                    dashboard_address = "http://%s:%d" % (
                        dashboard_host,
                        dashboard_port,
                    )
                    break
            else:
                dashboard_address = None

            with submit_and_handle_failures(skein_client, self.spec) as app:
                app.kv["dask.scheduler"] = scheduler_address.encode()
                if dashboard_address is not None:
                    app.kv["dask.dashboard"] = dashboard_address.encode()
        else:
            # deploy_mode == 'remote'
            with submit_and_handle_failures(skein_client, self.spec) as app:
                scheduler_address = app.kv.wait("dask.scheduler").decode()
                dashboard_address = app.kv.get("dask.dashboard")
                if dashboard_address is not None:
                    dashboard_address = dashboard_address.decode()

        # Ensure application gets cleaned up
        self._finalizer = weakref.finalize(self, app.shutdown)

        self.scheduler_address = scheduler_address
        self._dashboard_address = dashboard_address
        self.application_client = app

    def _connect_existing(self):
        spec = self.application_client.get_specification()
        if "dask.worker" not in spec.services:
            raise ValueError("%r is not a valid dask cluster" % self.app_id)

        scheduler_address = self.application_client.kv.wait("dask.scheduler").decode()
        dashboard_address = self.application_client.kv.get("dask.dashboard")
        if dashboard_address is not None:
            dashboard_address = dashboard_address.decode()

        self.spec = spec
        self.scheduler_address = scheduler_address
        self._dashboard_address = dashboard_address

    async def _start_internal(self):
        if self._start_task is None:
            self._start_task = asyncio.ensure_future(self._start_async())
        try:
            await self._start_task
        except BaseException:
            # On exception, cleanup
            await self._stop_internal()
            raise
        return self

    async def _start_async(self):
        if self.spec is not None:
            # Start a new cluster
            if "dask.scheduler" not in self.spec.services:
                self._scheduler = Scheduler(
                    loop=self.loop,
                    **self._scheduler_kwargs,
                )
                await self._scheduler
            else:
                self._scheduler = None
            await self.loop.run_in_executor(None, self._start_cluster)
        else:
            # Connect to an existing cluster
            await self.loop.run_in_executor(None, self._connect_existing)

        self.scheduler_comm = rpc(self.scheduler_address)
        comm = None
        try:
            comm = await self.scheduler_comm.live_comm()
            await comm.write({"op": "subscribe_worker_status"})
            self.scheduler_info = await comm.read()
            workers = self.scheduler_info.get("workers", {})
            self._requested.update(w["name"] for w in workers.values())
            self._watch_worker_status_task = asyncio.ensure_future(
                self._watch_worker_status(comm)
            )
        except Exception:
            if comm is not None:
                await comm.close()

    async def _stop_internal(self, status="SUCCEEDED", diagnostics=None):
        if self._stop_task is None:
            self._stop_task = asyncio.ensure_future(
                self._stop_async(status=status, diagnostics=diagnostics)
            )
        await self._stop_task

    async def _stop_async(self, status="SUCCEEDED", diagnostics=None):
        if self._start_task is not None:
            if not self._start_task.done():
                # We're still starting, cancel task
                await cancel_task(self._start_task)
            self._start_task = None

        if self._adaptive is not None:
            self._adaptive.stop()

        if self._watch_worker_status_task is not None:
            await cancel_task(self._watch_worker_status_task)
            self._watch_worker_status_task = None

        if self.scheduler_comm is not None:
            self.scheduler_comm.close_rpc()
            self.scheduler_comm = None

        await self.loop.run_in_executor(
            None, lambda: self._stop_sync(status=status, diagnostics=diagnostics)
        )

        if self._scheduler is not None:
            await self._scheduler.close()
            self._scheduler = None

    def _stop_sync(self, status="SUCCEEDED", diagnostics=None):
        if self._finalizer is not None and self._finalizer.peek() is not None:
            self.application_client.shutdown(status=status, diagnostics=diagnostics)
            self._finalizer.detach()  # don't run the finalizer later
        self._finalizer = None

    def __await__(self):
        return self.__aenter__().__await__()

    async def __aenter__(self):
        return await self._start_internal()

    async def __aexit__(self, typ, value, traceback):
        await self._stop_internal()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __repr__(self):
        return "YarnCluster<%s>" % self.app_id

    @property
    def loop(self):
        return self._loop_runner.loop

    @property
    def app_id(self):
        return self.application_client.id

    @property
    def asynchronous(self):
        return self._asynchronous

    def _sync(self, task):
        if self.asynchronous:
            return task
        future = asyncio.run_coroutine_threadsafe(task, self.loop.asyncio_loop)
        try:
            return future.result()
        except BaseException:
            future.cancel()
            raise

    @cached_property
    def dashboard_link(self):
        """Link to the dask dashboard. None if dashboard isn't running"""
        if self._dashboard_address is None:
            return None
        dashboard = urlparse(self._dashboard_address)
        return format_dashboard_link(dashboard.hostname, dashboard.port)

    def shutdown(self, status="SUCCEEDED", diagnostics=None):
        """Shutdown the application.

        Parameters
        ----------
        status : {'SUCCEEDED', 'FAILED', 'KILLED'}, optional
            The yarn application exit status.
        diagnostics : str, optional
            The application exit message, usually used for diagnosing failures.
            Can be seen in the YARN Web UI for completed applications under
            "diagnostics". If not provided, a default will be used.
        """
        if self.asynchronous:
            return self._stop_internal(status=status, diagnostics=diagnostics)
        if self.loop.asyncio_loop.is_running() and not sys.is_finalizing():
            self._sync(self._stop_internal(status=status, diagnostics=diagnostics))
        else:
            # Always run this!
            self._stop_sync(status=status, diagnostics=diagnostics)
        self._loop_runner.stop()

    def close(self, **kwargs):
        """Close this cluster. An alias for ``shutdown``.

        See Also
        --------
        shutdown
        """
        return self.shutdown(**kwargs)

    def __del__(self):
        if not hasattr(self, "_loop_runner"):
            return
        if self.asynchronous:
            # No del for async mode
            return
        self.close()

    @property
    def _observed(self):
        return {w["name"] for w in self.scheduler_info["workers"].values()}

    async def _workers(self):
        return await self.loop.run_in_executor(
            None,
            lambda: self.application_client.get_containers(services=["dask.worker"]),
        )

    def workers(self):
        """A list of all currently running worker containers."""
        return self._sync(self._workers())

    async def _logs(self, scheduler=True, workers=True):
        logs = Logs()

        if scheduler:
            slogs = await self.scheduler_comm.logs()
            logs["Scheduler"] = Log("\n".join(line for level, line in slogs))

        if workers:
            d = await self.scheduler_comm.worker_logs(workers=workers)
            for k, v in d.items():
                logs[k] = Log("\n".join(line for level, line in v))

        return logs

    def logs(self, scheduler=True, workers=True):
        """Return logs for the scheduler and/or workers

        Parameters
        ----------
        scheduler : boolean, optional
            Whether or not to collect logs for the scheduler
        workers : boolean or iterable, optional
            A list of worker addresses to select. Defaults to all workers if
            ``True`` or no workers if ``False``

        Returns
        -------
        logs : dict
            A dictionary of name -> logs.
        """
        return self._sync(self._logs(scheduler=scheduler, workers=workers))

    async def _scale_up(self, n):
        if n > len(self._requested):
            containers = await self.loop.run_in_executor(
                None,
                lambda: self.application_client.scale("dask.worker", n),
            )
            self._requested.update(c.id for c in containers)

    async def _scale_down(self, workers):
        self._requested.difference_update(workers)
        await self.scheduler_comm.retire_workers(names=list(workers))

        def _kill_containers():
            for c in workers:
                try:
                    self.application_client.kill_container(c)
                except ValueError:
                    pass

        await self.loop.run_in_executor(
            None,
            _kill_containers,
        )

    async def _scale(self, n):
        if self._adaptive is not None:
            self._adaptive.stop()
        if n >= len(self._requested):
            return await self._scale_up(n)
        else:
            n_to_delete = len(self._requested) - n
            pending = list(self._requested - self._observed)
            running = list(self._observed.difference(pending))
            to_close = pending[:n_to_delete]
            to_close.extend(running[: n_to_delete - len(to_close)])
            await self._scale_down(to_close)
            self._requested.difference_update(to_close)

    def scale(self, n):
        """Scale cluster to n workers.

        Parameters
        ----------
        n : int
            Target number of workers

        Examples
        --------
        >>> cluster.scale(10)  # scale cluster to ten workers
        """
        return self._sync(self._scale(n))

    def adapt(
        self,
        minimum=0,
        maximum=math.inf,
        interval="1s",
        wait_count=3,
        target_duration="5s",
        **kwargs
    ):
        """Turn on adaptivity

        This scales Dask clusters automatically based on scheduler activity.

        Parameters
        ----------
        minimum : int, optional
            Minimum number of workers. Defaults to ``0``.
        maximum : int, optional
            Maximum number of workers. Defaults to ``inf``.
        interval : timedelta or str, optional
            Time between worker add/remove recommendations.
        wait_count : int, optional
            Number of consecutive times that a worker should be suggested for
            removal before we remove it.
        target_duration : timedelta or str, optional
            Amount of time we want a computation to take. This affects how
            aggressively we scale up.
        **kwargs :
            Additional parameters to pass to
            ``distributed.Scheduler.workers_to_close``.

        Examples
        --------
        >>> cluster.adapt(minimum=0, maximum=10)
        """
        if self._adaptive is not None:
            self._adaptive.stop()
        self._adaptive_options.update(
            minimum=minimum,
            maximum=maximum,
            interval=interval,
            wait_count=wait_count,
            target_duration=target_duration,
            **kwargs,
        )
        self._adaptive = Adaptive(self, **self._adaptive_options)

    async def _watch_worker_status(self, comm):
        # We don't want to hold on to a ref to self, otherwise this will
        # leave a dangling reference and prevent garbage collection.
        ref_self = weakref.ref(self)
        self = None
        try:
            while True:
                try:
                    msgs = await comm.read()
                except OSError:
                    break
                try:
                    self = ref_self()
                    if self is None:
                        break
                    for op, msg in msgs:
                        if op == "add":
                            workers = msg.pop("workers")
                            self.scheduler_info["workers"].update(workers)
                            self.scheduler_info.update(msg)
                            self._requested.update(w["name"] for w in workers.values())
                        elif op == "remove":
                            del self.scheduler_info["workers"][msg]
                            self._requested.discard(msg)
                    if hasattr(self, "_status_widget"):
                        self._status_widget.value = self._widget_status()
                finally:
                    self = None
        finally:
            await comm.close()

    def _widget_status(self):
        try:
            workers = self.scheduler_info["workers"]
        except KeyError:
            return None
        else:
            n_workers = len(workers)
            cores = sum(w["nthreads"] for w in workers.values())
            memory = sum(w["memory_limit"] for w in workers.values())

            return _widget_status_template % (n_workers, cores, format_bytes(memory))

    def _widget(self):
        """ Create IPython widget for display within a notebook """
        try:
            return self._cached_widget
        except AttributeError:
            pass

        if self.asynchronous:
            return None

        try:
            from ipywidgets import Layout, VBox, HBox, IntText, Button, HTML, Accordion
        except ImportError:
            self._cached_widget = None
            return None

        layout = Layout(width="150px")

        title = HTML("<h2>YarnCluster</h2>")

        status = HTML(self._widget_status(), layout=Layout(min_width="150px"))

        request = IntText(0, description="Workers", layout=layout)
        scale = Button(description="Scale", layout=layout)

        minimum = IntText(0, description="Minimum", layout=layout)
        maximum = IntText(0, description="Maximum", layout=layout)
        adapt = Button(description="Adapt", layout=layout)

        accordion = Accordion(
            [HBox([request, scale]), HBox([minimum, maximum, adapt])],
            layout=Layout(min_width="500px"),
        )
        accordion.selected_index = None
        accordion.set_title(0, "Manual Scaling")
        accordion.set_title(1, "Adaptive Scaling")

        @adapt.on_click
        def adapt_cb(b):
            self.adapt(minimum=minimum.value, maximum=maximum.value)

        @scale.on_click
        def scale_cb(b):
            with log_errors():
                self.scale(request.value)

        app_id = HTML("<p><b>Application ID: </b>{0}</p>".format(self.app_id))

        elements = [title, HBox([status, accordion]), app_id]

        if self.dashboard_link is not None:
            link = HTML(
                '<p><b>Dashboard: </b><a href="{0}" target="_blank">{0}'
                "</a></p>\n".format(self.dashboard_link)
            )
            elements.append(link)

        self._cached_widget = box = VBox(elements)
        self._status_widget = status

        return box

    def _ipython_display_(self, **kwargs):
        widget = self._widget()
        if widget is not None:
            return widget._ipython_display_(**kwargs)
        else:
            from IPython.display import display

            data = {"text/plain": repr(self), "text/html": self._repr_html_()}
            display(data, raw=True)

    def _repr_html_(self):
        if self.dashboard_link is not None:
            dashboard = "<a href='{0}' target='_blank'>{0}</a>".format(
                self.dashboard_link
            )
        else:
            dashboard = "Not Available"
        return (
            "<div style='background-color: #f2f2f2; display: inline-block; "
            "padding: 10px; border: 1px solid #999999;'>\n"
            "  <h3>YarnCluster</h3>\n"
            "  <ul>\n"
            "    <li><b>Application ID: </b>{app_id}\n"
            "    <li><b>Dashboard: </b>{dashboard}\n"
            "  </ul>\n"
            "</div>\n"
        ).format(app_id=self.app_id, dashboard=dashboard)


class Adaptive(AdaptiveCore):
    def __init__(
        self,
        cluster=None,
        interval="1s",
        minimum=0,
        maximum=math.inf,
        wait_count=3,
        target_duration="5s",
        **kwargs
    ):
        self.cluster = cluster
        self.target_duration = parse_timedelta(target_duration)
        self._workers_to_close_kwargs = kwargs

        super().__init__(
            minimum=minimum, maximum=maximum, wait_count=wait_count, interval=interval
        )

    @property
    def requested(self):
        return self.cluster._requested

    plan = requested

    @property
    def observed(self):
        return self.cluster._observed

    async def target(self):
        return await self.cluster.scheduler_comm.adaptive_target(
            target_duration=self.target_duration
        )

    async def workers_to_close(self, target):
        return await self.cluster.scheduler_comm.workers_to_close(
            target=target, attribute="name", **self._workers_to_close_kwargs
        )

    async def scale_down(self, workers):
        await self.cluster._scale_down(workers)

    async def scale_up(self, n):
        await self.cluster._scale_up(n)

    @property
    def loop(self):
        return self.cluster.loop



from setuptools import setup
import versioneer

with open("requirements.txt") as f:
    install_requires = f.read().strip().split("\n")

with open("README.rst") as f:
    long_description = f.read()

setup(
    name="dask-yarn",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license="BSD",
    description="Deploy dask clusters on Apache YARN",
    long_description=long_description,
    packages=["dask_yarn"],
    include_package_data=True,
    install_requires=install_requires,
    python_requires=">=3.7",
    entry_points="""
        [console_scripts]
        dask-yarn=dask_yarn.cli:main
      """,
    zip_safe=False,
)



import os
import subprocess
import sys
import time
from contextlib import contextmanager

import skein
import pytest


@pytest.fixture(scope="session")
def conda_env():
    envpath = "dask-yarn-py%d%d.tar.gz" % sys.version_info[:2]
    if not os.path.exists(envpath):
        conda_pack = pytest.importorskip("conda_pack")
        conda_pack.pack(output=envpath, verbose=True)
    return envpath


@pytest.fixture(scope="session")
def skein_client():
    with skein.Client() as client:
        yield client


def check_is_shutdown(client, app_id, status="SUCCEEDED"):
    timeleft = 10
    report = client.application_report(app_id)
    while report.state not in ("FINISHED", "FAILED", "KILLED"):
        time.sleep(0.1)
        timeleft -= 0.1
        report = client.application_report(app_id)
        if timeleft < 0:
            client.kill_application(app_id)
            logs = get_logs(app_id)
            print(
                "Application wasn't properly terminated, killed by test fixture.\n"
                "\n"
                "Application Logs\n"
                "----------------\n"
                "%s" % logs
            )
            assert False, "Application wasn't properly terminated"

    if report.final_status != status:
        logs = get_logs(app_id)
        print(
            "Expected application to terminate with status==%s, got status==%s\n"
            "\n"
            "Application Logs\n"
            "----------------\n"
            "%s" % (status, report.final_status, logs)
        )
        assert report.final_status == status


@contextmanager
def ensure_shutdown(client, app_id, status=None):
    try:
        yield
    except Exception:
        client.kill_application(app_id)
        raise
    else:
        try:
            final_status = wait_for_completion(client, app_id, timeout=5)
        except AssertionError:
            client.kill_application(app_id)
            raise
        assert final_status == status


def wait_for_completion(client, app_id, timeout=30):
    while timeout:
        final_status = client.application_report(app_id).final_status
        if final_status != "UNDEFINED":
            return final_status
        time.sleep(0.1)
        timeout -= 0.1
    else:
        assert False, "Application timed out"


def get_logs(app_id, tries=3):
    command = ["yarn", "logs", "-applicationId", app_id]
    for i in range(tries - 1):
        try:
            return subprocess.check_output(command).decode()
        except Exception:
            pass
        time.sleep(1)
    return subprocess.check_output(command).decode()
